package fr.sle;

import org.drools.core.ClockType;
import org.drools.core.RuleBaseConfiguration;
import org.drools.core.SessionConfiguration;
import org.drools.core.event.DebugRuleRuntimeEventListener;
import org.drools.core.marshalling.impl.IdentityPlaceholderResolverStrategy;
import org.drools.core.time.SessionPseudoClock;
import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;
import org.kie.api.KieBase;
import org.kie.api.KieBaseConfiguration;
import org.kie.api.conf.EqualityBehaviorOption;
import org.kie.api.conf.EventProcessingOption;
import org.kie.api.event.rule.DebugAgendaEventListener;
import org.kie.api.io.ResourceType;
import org.kie.api.marshalling.Marshaller;
import org.kie.api.marshalling.ObjectMarshallingStrategy;
import org.kie.api.runtime.KieSession;
import org.kie.api.runtime.KieSessionConfiguration;
import org.kie.api.runtime.conf.ClockTypeOption;
import org.kie.internal.marshalling.MarshallerFactory;
import org.kie.internal.utils.KieHelper;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.time.Duration;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.concurrent.TimeUnit;

@RunWith(JUnit4.class)
public class ReproducerTest {

    private String declareDrl =
            "import fr.sle.AEvent;\n" +
                    "\n" +
                    "declare AEvent\n" +
                    "    @role(event)\n" +
                    "    @timestamp(eventTimestamp)\n" +
                    "    @expires (3d)\n" +
                    "end";

    private String ruleDrl_1 =
            "rule \"set value\"\n" +
                    "no-loop \n" +
                    "when\n" +
                    "    $p: AEvent()\n" +
                    "then\n" +
                    "    modify($p){\n" +
                    "        setValue(\"value\")\n" +
                    "    }\n" +
                    "end";

    private String ruleDrl_2 =
            "import fr.sle.BEvent;\n" +
                    "rule \"right tuple rule\"\n" +
                    "when\n" +
                    "    $b: BEvent()\n" +
                    "    $a: AEvent()\n" +
                    "then\n" +
                    "    System.out.println(\"right tuple rule\");\n" +
                    "end";


    @Test
    public void testAvoidAlreadyExpiredFactsToForeverRemainInWM() throws IOException, ClassNotFoundException {
        final KieHelper helper = new KieHelper();
        helper.addContent(declareDrl, ResourceType.DRL);
        helper.addContent(ruleDrl_1, ResourceType.DRL);
        helper.addContent(ruleDrl_2, ResourceType.DRL);

        KieBaseConfiguration kieBaseConfiguration = new RuleBaseConfiguration();
        kieBaseConfiguration.setOption(EqualityBehaviorOption.EQUALITY);
        kieBaseConfiguration.setOption(EventProcessingOption.STREAM);
        final KieBase kieBase = helper.build(kieBaseConfiguration);

        KieSessionConfiguration kieSessionConfiguration = SessionConfiguration.newInstance();
        kieSessionConfiguration.setOption(ClockTypeOption.get(ClockType.PSEUDO_CLOCK.getId()));

        final KieSession kieSession = kieBase.newKieSession(kieSessionConfiguration, null);

        final long now = System.currentTimeMillis();
        final SessionPseudoClock clock = kieSession.getSessionClock();
        clock.advanceTime(now, TimeUnit.MILLISECONDS);
        kieSession.fireAllRules();

        kieSession.addEventListener(new DebugAgendaEventListener());
        kieSession.addEventListener(new DebugRuleRuntimeEventListener());

        final Date nowMinus5d = new Date(now - Duration.of(5, ChronoUnit.DAYS).toMillis());
        final AEvent aEvent = new AEvent("001A", nowMinus5d);

        kieSession.insert(aEvent);
        final int r1 = kieSession.fireAllRules();

        Assert.assertEquals(1, r1);

		//clock.advanceTime(now + Duration.of(1,ChronoUnit.DAYS).toMillis(), TimeUnit.MILLISECONDS);   // for drools 6.5.x
        //kieSession.fireAllRules();																   // for drools 6.5.x

        Assert.assertEquals(0, kieSession.getFactCount()); //Problem: AEvent still in WM

		//Comment to previous assert to show some bad consequences
        kieSession.insert(new BEvent("002B", new Date(now)));

		final int r2 = kieSession.fireAllRules();
        System.out.println("Number of fired rules : " + r2); // should be 0, No match with BEvent, normal --> RightTuple link to AEvent is expired

        final AEvent aClonedEvent = new AEvent("001A", new Date(now));
        kieSession.insert(aClonedEvent);
        System.out.println("Number of facts in WM: " + kieSession.getFactCount()); //We can no longer add AEvent, because it lives forever in working memory

        final IdentityPlaceholderResolverStrategy identityPlaceholderResolverStrategy = new IdentityPlaceholderResolverStrategy((Object o) -> true);

        byte[] serialized = null;
        final Marshaller marshaller = MarshallerFactory.newMarshaller(kieBase, new ObjectMarshallingStrategy[]{ identityPlaceholderResolverStrategy });
        try (ByteArrayOutputStream baos = new ByteArrayOutputStream()) {
            marshaller.marshall(baos, kieSession);
            serialized = baos.toByteArray();
        }

        KieSession unmarshallSession = null;
        try(ByteArrayInputStream bais = new ByteArrayInputStream(serialized)){
            final Marshaller unmarshaller = MarshallerFactory.newMarshaller(kieBase, new ObjectMarshallingStrategy[]{ identityPlaceholderResolverStrategy });
            unmarshallSession = unmarshaller.unmarshall(bais);
        }

        final int r3 = unmarshallSession.fireAllRules();

        System.err.println("Number of triggers: " + r3);

    }


}
